# Copyright (c) 2023, Djaodjin Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging, os

from django.db import connections
from django.db.utils import DEFAULT_DB_ALIAS
from django.utils.encoding import iri_to_uri

from . import settings
from .compat import python_2_unicode_compatible, reverse, urljoin, urlparse

try:
    from threading import local
except ImportError:
    #pylint: disable=import-error,no-name-in-module
    from django.utils._threading_local import local

_thread_locals = local() #pylint: disable=invalid-name

LOGGER = logging.getLogger(__name__)


@python_2_unicode_compatible
class CurrentSite(object):

    def __init__(self, site, path_prefix,
                 default_scheme='http', default_host='localhost'):
        self.db_object = site
        self.path_prefix = path_prefix
        self.default_scheme = default_scheme
        self.default_host = default_host

    def __getattr__(self, name):
        return getattr(self.db_object, name)

    def __str__(self):
        return self.db_object.__str__()

    def as_absolute_uri(self, location='/'):
        parts = urlparse(location)
        if parts.scheme and parts.netloc:
            # If we already have an absolute URI then we have not processing
            # to do and just return it "as is".
            return location

        if self.db_object.domain:
            host = self.db_object.domain
        else:
            host = self.default_host
            if self.path_prefix:
                location = location.lstrip('/')
                if location.startswith(self.path_prefix):
                    location = location[len(self.path_prefix):]
                # `lstrip` again In case we removed the path_prefix previously.
                location = urljoin('/%s/' % self.path_prefix,
                    location.lstrip('/'))
        return iri_to_uri('%(scheme)s://%(host)s%(path)s' % {
            'scheme': self.default_scheme, 'host': host, 'path': location})


def as_provider_db(db_name, db_host=None, db_port=None):
    """
    Returns a dictionnary that can be used to initialized a database
    connection to the a site-specific database.
    """
    provider_db = connections.databases[DEFAULT_DB_ALIAS].copy()
    default_db_name = provider_db['NAME']
    provider_db.update({'NAME':db_name})
    if db_host:
        provider_db.update({'HOST':db_host})
    if db_port:
        provider_db.update({'PORT':db_port})
    if provider_db['ENGINE'].endswith('sqlite3'):
        # HACK to set absolute paths (used in development environments)
        candidates = [os.path.join(dir_path, db_name + '.sqlite')
            for dir_path in [os.path.dirname(default_db_name)]
                       + settings.DEBUG_SQLITE3_PATHS]
        # When the sqlite database file does not yet exists, we want
        # to default to a db file in the same directory as default,
        # not a file in the current directory.
        provider_db['NAME'] = candidates[0]
        for candidate_db in candidates:
            if os.path.exists(candidate_db):
                provider_db['NAME'] = candidate_db
                LOGGER.debug("multitier: using database '%s'", candidate_db)
                return provider_db
        LOGGER.error("multitier: cannot find db '%s'", db_name)
    return provider_db


def cache_provider_db(db_name, db_host=None, db_port=None):
    if not db_name:
        return None
    if not db_name in connections.databases:
        connections.databases[db_name] = as_provider_db(db_name,
            db_host=db_host, db_port=db_port)
    return connections.databases[db_name]


def clear_cache():
    try:
        del _thread_locals.site
    except AttributeError:
        pass


def get_current_site():
    """
    Returns the ``Site`` associated to the thread request.
    """
    return getattr(_thread_locals, 'site', None)


def get_path_prefix():
    """
    Returns the prefix every URL paths is prefixed with.
    """
    site = get_current_site()
    if site:
        return site.path_prefix
    return ""


def set_current_site(site, path_prefix,
        default_scheme='http', default_host='localhost', request=None):
    # Dynamically update the db used for auth and saas.
    if site.db_name:
        LOGGER.debug(
            "multitier: access site '%s' with prefix '%s', connect to db '%s'",
            site, path_prefix, site.db_name)
        if not site.db_name in connections.databases:
            cache_provider_db(site.db_name,
                db_host=site.db_host, db_port=site.db_port)
    else:
        LOGGER.debug(
            "multitier: access site '%s' with prefix '%s',"\
            " connect to db 'default'", site, path_prefix)

    prev_site = None
    prev_path_prefix = None
    if (hasattr(_thread_locals, 'site')
        and isinstance(_thread_locals.site, CurrentSite)):
        prev_site = _thread_locals.site.db_object
        prev_path_prefix = _thread_locals.site.path_prefix
        _thread_locals.site.db_object = site
        _thread_locals.site.path_prefix = path_prefix
    else:
        _thread_locals.site = CurrentSite(site, path_prefix,
            default_scheme=default_scheme, default_host=default_host)

    if request is not None:
        request.site = _thread_locals.site
        request.urls = {}
        for url_name in settings.DEFAULT_URLS:
            request.urls.update({url_name: reverse(url_name)})
    return (prev_site, prev_path_prefix)



# Copyright (c) 2021, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#pylint: disable=unused-import,bad-except-order
#pylint:disable=no-name-in-module,import-error,import-outside-toplevel
import six

from six.moves.urllib.parse import urlparse, urlunparse, urljoin

try:
    from functools import lru_cache
except ImportError: # python < 3
    from django.utils.lru_cache import lru_cache


try:
    from django.template import Origin
except ImportError:
    Origin = None


try:
    from django.urls import (NoReverseMatch, URLPattern as RegexURLPattern,
        URLResolver as RegexURLResolver, reverse, reverse_lazy)
except ImportError: # <= Django 1.10, Python<3.6
    from django.core.urlresolvers import (NoReverseMatch, RegexURLPattern,
        RegexURLResolver, reverse, reverse_lazy)
except ModuleNotFoundError: #pylint:disable=undefined-variable
    # <= Django 1.10, Python>=3.6
    from django.core.urlresolvers import (NoReverseMatch, RegexURLPattern,
        RegexURLResolver, reverse, reverse_lazy)

try:
    from django.urls import include, re_path
except ImportError: # <= Django 2.0, Python<3.6
    from django.conf.urls import include, url as re_path


try:
    from django.utils.module_loading import import_string
except ImportError: # django < 1.7
    from django.utils.module_loading import import_by_path as import_string


try:
    from django.utils.encoding import python_2_unicode_compatible
except ImportError: # django < 3.0
    python_2_unicode_compatible = six.python_2_unicode_compatible


try:
    from django.utils.translation import gettext_lazy
except ImportError: # django < 3.0
    from django.utils.translation import ugettext_lazy as gettext_lazy


def get_model_class(full_name, settings_meta=None):
    """
    Returns a model class loaded from *full_name*. *settings_meta* is the name
    of the corresponding settings variable (used for error messages).
    """
    from django.core.exceptions import ImproperlyConfigured

    try:
        app_label, model_name = full_name.split('.')
    except ValueError:
        raise ImproperlyConfigured(
            "%s must be of the form 'app_label.model_name'" % full_name)
    return get_app_model_class(app_label, model_name, settings_meta)


def get_app_model_class(app_label, model_name, settings_meta=None):
    """
    Returns a model class loaded from *full_name*. *settings_meta* is the name
    of the corresponding settings variable (used for error messages).
    """
    from django.core.exceptions import ImproperlyConfigured
    model_class = None
    try:
        from django.apps import apps
        model_class = apps.get_model(app_label, model_name)
    except ImportError: # django < 1.7
        #pylint: disable=no-name-in-module
        from django.db.models import get_model
        model_class = get_model(app_label, model_name)

    if model_class is None:
        if settings_meta is None:
            settings_meta = 'settings'
        raise ImproperlyConfigured(
            "%s refers to model '%s' that has not been installed"
            % (settings_meta, app_label + '.' + model_name))
    return model_class


try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError: # django < 1.11
    class MiddlewareMixin(object):
        pass



# Copyright (c) 2019, Djaodjin Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from django.urls.resolvers import URLPattern, URLResolver
from django.utils.datastructures import MultiValueDict
from django.utils.regex_helper import normalize


from .thread_locals import get_current_site


class RegexURLResolver(URLResolver):
    """
    A URL resolver that always matches the active organization code
    as URL prefix.
    """
    def __init__(self, regex, urlconf_name, *args, **kwargs):
        super(RegexURLResolver, self).__init__(
            regex, urlconf_name, *args, **kwargs)

    @staticmethod
    def _get_path_prefix():
        path_prefix = "_"
        current_site = get_current_site()
        if current_site and current_site.path_prefix:
            path_prefix = current_site.path_prefix
        return path_prefix

    # Implementation Note:
    # Copy/Pasted `RegexURLResolver._populate` here because that was the only
    # way to override `language_code = get_language()` to use a dynamic path
    # prefix `path_prefix = self._get_path_prefix()`.
    def _populate(self):
        # Short-circuit if called recursively in this thread to prevent
        # infinite recursion. Concurrent threads may call this at the same
        # time and will need to continue, so set 'populating' on a
        # thread-local variable.
        if getattr(self._local, 'populating', False):
            return
        try:
            self._local.populating = True
            lookups = MultiValueDict()
            namespaces = {}
            apps = {}
            path_prefix = self._get_path_prefix()
            for url_pattern in reversed(self.url_patterns):
                p_pattern = url_pattern.pattern.regex.pattern
                if p_pattern.startswith('^'):
                    p_pattern = p_pattern[1:]
                if isinstance(url_pattern, URLPattern):
                    self._callback_strs.add(url_pattern.lookup_str)
                    bits = normalize(url_pattern.pattern.regex.pattern)
                    lookups.appendlist(
                        url_pattern.callback, (
                            bits, p_pattern,
                            url_pattern.default_args,
                            url_pattern.pattern.converters)
                    )
                    if url_pattern.name is not None:
                        lookups.appendlist(url_pattern.name, (
                            bits, p_pattern,
                            url_pattern.default_args,
                            url_pattern.pattern.converters)
                        )
                else:  # url_pattern is a URLResolver.
                    url_pattern._populate()
                    if url_pattern.app_name:
                        apps.setdefault(url_pattern.app_name, []).append(
                            url_pattern.namespace)
                        namespaces[url_pattern.namespace] = (
                            p_pattern, url_pattern)
                    else:
                        for name in url_pattern.reverse_dict:
                            for matches, pat, defaults, converters in \
                                    url_pattern.reverse_dict.getlist(name):
                                new_matches = normalize(p_pattern + pat)
                                lookups.appendlist(
                                    name,
                                    (
                                        new_matches,
                                        p_pattern + pat,
                                        {**defaults,
                                         **url_pattern.default_kwargs},
                                        {**self.pattern.converters,
                                         **url_pattern.pattern.converters,
                                         **converters}
                                    )
                                )
                        for namespace, (prefix, sub_pattern) in \
                                url_pattern.namespace_dict.items():
                            current_converters = \
                                url_pattern.pattern.converters
                            sub_pattern.pattern.converters.update(
                                current_converters)
                            namespaces[namespace] = (
                                p_pattern + prefix, sub_pattern)
                        for app_name, namespace_list in \
                                url_pattern.app_dict.items():
                            apps.setdefault(app_name, []).extend(
                                namespace_list)
                    self._callback_strs.update(url_pattern._callback_strs)
            self._namespace_dict[path_prefix] = namespaces
            self._app_dict[path_prefix] = apps
            self._reverse_dict[path_prefix] = lookups
            self._populated = True
        finally:
            self._local.populating = False

    @property
    def reverse_dict(self):
        path_prefix = self._get_path_prefix()
        if path_prefix not in self._reverse_dict:
            self._populate()
        return self._reverse_dict[path_prefix]

    @property
    def namespace_dict(self):
        path_prefix = self._get_path_prefix()
        if path_prefix not in self._namespace_dict:
            self._populate()
        return self._namespace_dict[path_prefix]

    @property
    def app_dict(self):
        path_prefix = self._get_path_prefix()
        if path_prefix not in self._app_dict:
            self._populate()
        return self._app_dict[path_prefix]



# Copyright (c) 2020, Djaodjin Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from django.conf import settings as django_settings
from django.db import DEFAULT_DB_ALIAS

from . import settings
from .compat import get_app_model_class, six
from .thread_locals import get_current_site


class SiteRouter(object):
    """
    A router to control all database operations on Django models in a set
    of ``apps`` and SQL tables in a set of ``tables``.
    """
    #pylint: disable=protected-access

    apps = settings.ROUTER_APPS
    tables = settings.ROUTER_TABLES

    @staticmethod
    def provider_db():
        if hasattr(django_settings, 'MULTITIER_NAME'):
            # ``manage.py loaddata`` will call db_for_write for relation
            # tables. Since we don't have a ``request`` then, we rely
            # on an environment variable.
            return django_settings.MULTITIER_NAME
        current_site = get_current_site()
        if current_site:
            return current_site.db_name
        return DEFAULT_DB_ALIAS

    def includes(self, model):
        if isinstance(model, six.string_types):
            return model in self.apps
        return (model._meta.app_label in self.apps
                or model._meta.db_table in self.tables)


    def db_for_read(self, model, **hints): #pylint: disable=unused-argument
        """
        Attempts to read ``apps`` models go to the current provider.
        """
        result = None
        if self.includes(model):
            result = self.provider_db()
        return result

    def db_for_write(self, model, **hints): #pylint: disable=unused-argument
        """
        Attempts to write ``apps`` models to the current provider.
        """
        result = None
        if self.includes(model):
            result = self.provider_db()
        return result

    def allow_relation(self, obj1, obj2, **hints):
        #pylint: disable=unused-argument
        """
        Allow relations if a model in the ``apps`` is involved.
        """
        if obj1._meta.app_label in self.apps or \
                obj2._meta.app_label in self.apps:
            return True
        return None

    # XXX Note here running tests with Django <1.7 will call
    #     allow_syncdb instead.
    def allow_migrate(self, database, app_label, model_name=None, **hints):
        """
        Make sure the apps only appears in the current provider
        database.
        """
        result = True
        model = hints.get('model')
        if model is None:
            if model_name is not None:
                model = get_app_model_class(app_label, model_name)
            else:
                # Django 1.7 prototype is allow_migrate(self, db, model)
                model = app_label
        if database != DEFAULT_DB_ALIAS:
            result = self.includes(model)
        return result



# Copyright (c) 2017, Djaodjin Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import absolute_import

import logging

import django
from django.conf import settings
from django.template.loaders.filesystem import Loader as FilesystemLoader
from django.utils._os import safe_join

from multitier.thread_locals import get_current_site
from multitier.compat import Origin


LOGGER = logging.getLogger(__name__)


class Loader(FilesystemLoader):

    def __init__(self, engine):
        super(Loader, self).__init__(engine)
        self.encoding = 'utf-8'

    def searchpath(self, template_dirs=None):
        if not template_dirs:
            try:
                template_dirs = self.get_dirs() #pylint:disable=no-member
            except AttributeError: # django < 1.8
                template_dirs = settings.TEMPLATE_DIRS
        current_site = get_current_site()
        if current_site:
            loader_template_dirs = []
            for template_dir in current_site.get_template_dirs():
                loader_template_dirs.append(safe_join(template_dir, 'django'))
                loader_template_dirs.append(template_dir)
            template_dirs = loader_template_dirs + list(template_dirs)
        return template_dirs

    def get_template_sources(self, template_name, template_dirs=None):
        try:
            for template_dir in self.searchpath(template_dirs=template_dirs):
                try:
                    template_path = safe_join(template_dir, template_name)
                    if django.VERSION[0] <= 1 and django.VERSION[1] < 9:
                        yield template_path
                    else:
                        yield Origin(
                            name=template_path,
                            template_name=template_name,
                            loader=self)
                except UnicodeDecodeError:
                    # The template dir name was a bytestring that wasn't
                    # valid UTF-8.
                    raise
                except ValueError:
                    # The joined path was located outside
                    # of this particular template_dir (it might be
                    # inside another one, so this isn't fatal).
                    pass

        except AttributeError as attr_err:
            # Something bad appended. We don't even have a request.
            # The middleware might be misconfigured.
            LOGGER.warning(
                "%s, your middleware might be misconfigured.", attr_err)



# Copyright (c) 2019, Djaodjin Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import absolute_import

import hashlib, logging, os

from django.utils._os import safe_join
import jinja2

from multitier.thread_locals import get_current_site


LOGGER = logging.getLogger(__name__)


class Loader(jinja2.FileSystemLoader):
    """
    Jinja2 loader.
    """
    def __init__(self, searchpath, encoding='utf-8', followlinks=False):
        super(Loader, self).__init__(
            searchpath, encoding=encoding, followlinks=followlinks)

    def get_template_dirs(self, template_dirs=None):
        if template_dirs is None:
            template_dirs = self.searchpath
        current_site = get_current_site()
        if current_site:
            loader_template_dirs = []
            for template_dir in current_site.get_template_dirs():
                loader_template_dirs.append(safe_join(template_dir, 'jinja2'))
                loader_template_dirs.append(template_dir)
            template_dirs = loader_template_dirs + list(template_dirs)
        return template_dirs


    def get_source(self, environment, template):
        pieces = jinja2.loaders.split_template_path(template)
        filename = None
        contents = None
        digest = None
        for searchpath in self.get_template_dirs():
            filename = os.path.join(searchpath, *pieces)
            if os.path.isfile(filename):
                LOGGER.debug("found template %s", filename)
                with open(filename, "rb") as template_file:
                    data = template_file.read()
                    digest = hashlib.sha1(data).hexdigest()
                    contents = data.decode(self.encoding)
                break
#            else:
#                LOGGER.debug("tried template %s", filename)
        if filename is not None and contents is not None:
            def uptodate():
                try:
                    with open(filename, "rb") as template_file:
                        file_digest = hashlib.sha1(
                            template_file.read()).hexdigest()
                        return file_digest == digest
                except OSError:
                    return False
            return contents, filename, uptodate
        raise jinja2.exceptions.TemplateNotFound(template)



# Copyright (c) 2022, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from django import template
from django.templatetags.static import StaticNode

from ..compat import six, urljoin
from ..mixins import build_absolute_uri
from .. import settings
from ..thread_locals import get_current_site


register = template.Library()


class MultitierStaticNode(StaticNode):

    def url(self, context):
        path = super(MultitierStaticNode, self).url(context)
        return site_url(path)


@register.tag('static')
def do_static(parser, token):
    """
    A template tag that returns the URL to a file
    using a ``multitier.Site``

    Usage::

        {% static path [as varname] %}

    Examples::

        {% static "myapp/css/base.css" %}
        {% static variable_with_path %}
        {% static "myapp/css/base.css" as admin_base_css %}
        {% static variable_with_path as varname %}

    """
    return MultitierStaticNode.handle_token(parser, token)


@register.filter()
def asset(path):
    """
    Adds the appropriate url or path prefix.

    While ``{% static path [as varname] %}`` would work in the context
    of Django templates, ``{{ path|asset }}`` works in both Django and Jinja2
    templates.
    """
    return site_url(path)


@register.filter()
def site_printable_name(request):
    if hasattr(request, 'site'):
        return request.site.printable_name
    return settings.DEFAULT_SITE


@register.filter()
def site_url(request):
    if isinstance(request, six.string_types):
        path = request
        if path is None:
            path = ''
        site = get_current_site()
        if site and site.path_prefix:
            path_prefix = '/%s' % site.path_prefix
        else:
            path_prefix = ''
        if path:
            # We have an actual path instead of generating a prefix that will
            # be placed in front of static urls (ie. {{'pricing'|site_url}}
            # insted of {{''|site_url}}{{ASSET_URL}}).
            path_prefix += '/'
            path = path.lstrip('/')
        return urljoin(path_prefix, path)
    return build_absolute_uri(request).rstrip('/')



# Copyright (c) 2017, Djaodjin Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from collections import OrderedDict
import errno, os

from django.conf import settings as django_settings
from django.contrib.staticfiles.finders import (
    FileSystemFinder, AppDirectoriesFinder as BaseAppDirectoriesFinder)
from django.core.files.storage import FileSystemStorage
from django.contrib.staticfiles import utils

from .thread_locals import get_current_site
from .settings import STATICFILES_DIRS


def get_current_theme_assets_dirs():
    """
    Returns a path which is the directory root of static assets for a theme.
    """
    roots = []
    site = get_current_site()
    if site is not None:
        # ``site`` could be ``None`` when this code is used through
        # a manage.py command (ex: collectstatic).
        #
        # Here we are inserting the *theme* at a natural place,
        # i.e. before the path postfix matching STATIC_URL.
        url_parts = []
        for part in django_settings.STATIC_URL.split('/'):
            if part:
                url_parts.append(part)
        for static_dir in STATICFILES_DIRS:
            drive, path = os.path.splitdrive(static_dir)
            dir_parts = path.split(os.sep)
            nb_dir_parts = len(dir_parts)
            nb_url_parts = len(url_parts)
            cut_point = nb_dir_parts - nb_url_parts
            if cut_point > 0:
                for dir_part, url_part in zip(
                    dir_parts[cut_point:], url_parts):
                    if dir_part != url_part:
                        cut_point = nb_dir_parts
                        break
            else:
                cut_point = nb_dir_parts
            for theme in site.get_templates():
                roots.append(os.path.join(drive, os.sep,
                    *(dir_parts[:cut_point] + [theme]
                      + dir_parts[cut_point:])))
    return roots


#pylint:disable=no-member
class MultitierFileSystemFinder(FileSystemFinder):
    """
    A static files finder that uses ``get_current_site()`` to locate files.
    """

    def get_locations(self):
        #pylint:disable=too-many-locals
        locations = []
        storages = OrderedDict()
        roots = get_current_theme_assets_dirs()
        for root in roots:
            prefix = ''
            if (prefix, root) not in locations:
                locations.append((prefix, root))
                filesystem_storage = FileSystemStorage(location=root)
                filesystem_storage.prefix = prefix
                storages[root] = filesystem_storage
        locations = locations + self.locations
        storages.update(self.storages)
        return locations, storages

    def find(self, path, all=False): #pylint:disable=redefined-builtin
        """
        Looks for files in the extra locations
        as defined in ``STATICFILES_DIRS`` and multitier locations.
        """
        matches = []
        locations, _ = self.get_locations()
        for prefix, root in locations:
            matched_path = self.find_location(root, path, prefix)
            if matched_path:
                if not all:
                    return matched_path
                matches.append(matched_path)
        return matches

    def list(self, ignore_patterns):
        """
        List all files in all locations.
        """
        locations, storages = self.get_locations()
        for _, root in locations:
            storage = storages[root]
            try:
                for path in utils.get_files(storage, ignore_patterns):
                    yield path, storage
            except OSError as err:
                if err.errno == errno.ENOENT:
                    # suppress "No such file or directory" error because
                    # non-existent directories are considered optional search
                    # path here.
                    pass
                else:
                    raise


class AppDirectoriesFinder(BaseAppDirectoriesFinder):
    """
    Override of ``django.contrib.staticfiles.AppDirectoriesFinder`` that
    removes the source_dir prefix from path before continuing the search.
    It is the only way so far we found to run django_assets in both
    debug and production mode in the context of multitier sites.
    """

    def find_in_app(self, app, path):
        if path.startswith(self.source_dir):
            path = path[len(self.source_dir) + 1:]
        return super(AppDirectoriesFinder, self).find_in_app(app, path)



# Copyright (c) 2020, Djaodjin Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import logging, re

from django.conf import settings as django_settings
from django.db.models import Q
from django.http import Http404

from . import settings
from .utils import get_site_model
from .thread_locals import clear_cache, set_current_site
from .compat import MiddlewareMixin


LOGGER = logging.getLogger(__name__)


class SiteMiddleware(MiddlewareMixin):

    @staticmethod
    def as_candidate_site(request):
        """
        Returns a ``Site`` based on the request host.
        """
        site = None
        candidate = None
        path_prefix = ''
        all_host_allowed = False
        host = request.get_host().split(':')[0].lower()
        app_domain = host
        if django_settings.ALLOWED_HOSTS:
            first_allowed_host = django_settings.ALLOWED_HOSTS[0]
            if first_allowed_host == '*':
                all_host_allowed = True
            else:
                app_domain = django_settings.ALLOWED_HOSTS[0]
                if app_domain.startswith('.'):
                    app_domain = app_domain[1:]
            look = re.match(r'^((?P<subdomain>\S+)\.)?%s(?::.*)?$' % app_domain,
                host)
            if look and look.group('subdomain'):
                candidate = look.group('subdomain')
            if host == app_domain and not candidate:
                look = re.match(r'^/(?P<path_prefix>[a-zA-Z0-9\-]+).*',
                    request.path)
                # no trailing '/' is OK here.
                if look:
                    path_prefix = look.group('path_prefix')
                    # It is either a subdomain or a path_prefix. Trying both
                    # match one after the other will override the candidate.
                    if path_prefix:
                        candidate = path_prefix
        try:
            flt = Q(domain=host)
            if candidate:
                flt = flt | Q(slug=candidate)
            if all_host_allowed or host == app_domain:
                flt = flt | Q(slug=settings.DEFAULT_SITE)
            queryset = get_site_model().objects.filter(
                flt, is_active=True).order_by('-domain', '-pk')
            site = queryset.first()
            if site is None or (site.domain and site.domain != host):
                # We return a 404 if the site is accessed through
                # the default domain when a domain is present because
                # a `dig domain` will return the default domain.
                #pylint: disable=raising-bad-type
                raise get_site_model().DoesNotExist
            if not site.is_path_prefix or site.slug != path_prefix:
                path_prefix = ''
        except get_site_model().DoesNotExist:
            if candidate is not None:
                msg = "'%s' nor subdomain '%s%s' could be found." % (
                    host, candidate, django_settings.ALLOWED_HOSTS[0])
            else:
                msg = "'%s' could not be found." % str(host)
            LOGGER.debug(msg, extra={'request': request})
            raise Http404(msg)
        return site, path_prefix

    def process_request(self, request):
        """
        Adds a ``client`` attribute to the ``request`` parameter.
        """
        clear_cache()
        site, path_prefix = self.as_candidate_site(request)

        # This is where you would typically override ``request.urlconf``
        # based on the ``Site``.

        # Set thread locals.
        # 1. First the site such that the we route requests for ``Site``
        # instances to the correct database.
        # 2. Then we hack into the translation module to get
        # django.core.urlresolvers to play nicely with our url scheme
        # with regards to the active site.
        set_current_site(site, path_prefix,
            default_scheme=request.scheme, default_host=request.get_host(),
            request=request)


class SetRemoteAddrFromForwardedFor(MiddlewareMixin):
    """
    set REMOTE_ADDR based on HTTP_X_FORWARDED_FOR.
    """

    @staticmethod
    def process_request(request):
        if request.META.get('REMOTE_ADDR', '127.0.0.1') == '127.0.0.1':
            request.META.update({'REMOTE_ADDR':
                request.META.get('HTTP_X_FORWARDED_FOR', '127.0.0.1')})



# Copyright (c) 2022, Djaodjin Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Models for the multi-tier application.
"""

import json, re, string

from django.core.mail import get_connection as get_connection_base
from django.core.validators import (_lazy_re_compile, RegexValidator,
    URLValidator)
from django.core.exceptions import ValidationError
from django.db import models
from django.utils._os import safe_join

from deployutils.crypt import decrypt, encrypt

from . import settings
from .compat import gettext_lazy as _, python_2_unicode_compatible, six
from .utils import get_site_model


SUBDOMAIN_RE = r'^[-a-zA-Z0-9_]+\Z'
SUBDOMAIN_SLUG = RegexValidator(
    SUBDOMAIN_RE,
    _("Enter a valid subdomain consisting of letters, digits or hyphens."),
    'invalid'
)
HOST_VALIDATOR = RegexValidator(
    _lazy_re_compile(
        r'(?:' + URLValidator.ipv4_re + '|' + URLValidator.ipv6_re + \
        '|' + URLValidator.host_re + ')' # host
        r'(?::\d{2,5})?'                                       # port
        r'(?:[/?#][^\s]*)?'                                    # resource path
        r'\Z', re.IGNORECASE),
    _("Enter a valid host, optionally followed by a port and resource path."),
    'invalid'
)

def domain_name_validator(value):
    """
    Validates that the given value contains no whitespaces to prevent common
    typos.
    """
    if not value:
        return
    checks = ((s in value) for s in string.whitespace)
    if any(checks):
        raise ValidationError(
            _("The domain name cannot contain any spaces or tabs."),
            code='invalid',
        )


@python_2_unicode_compatible
class BaseSite(models.Model):

    # Since most DNS provider limit subdomain length to 25 characters,
    # we do here too. `django.contrib.Site` limits the domain length
    # to 100 characters so we also do the same. Though 100 characters
    # is often not enough for 3rd part db/email host names (ex: AWS RDS,
    # AWS SES) so we bumped the limit to 255 for those.
    slug = models.SlugField(unique=True, max_length=25,
        validators=[SUBDOMAIN_SLUG], help_text=_(
            "unique identifier for the site (also serves as subdomain)"))

    domain = models.CharField(max_length=100, null=True, blank=True,
        validators=[domain_name_validator, RegexValidator(
            URLValidator.host_re,
            _("Enter a valid 'domain', ex: example.com"), 'invalid')],
        help_text=_(
            _("fully qualified domain name at which the site is available")))
    is_path_prefix = models.BooleanField(default=False, help_text=_(
        "use slug as a prefix for URL paths instead of domain field."))
    cert_location = models.CharField(max_length=1024, null=True,
        help_text=_("Location of the TLS certificate for HTTPS connections"))

    account = models.ForeignKey(
        settings.ACCOUNT_MODEL, null=True, on_delete=models.CASCADE,
        related_name='sites')
    is_active = models.BooleanField(default=False,
        help_text=_("The Site is active or not"))
    extra = models.CharField(null=True, max_length=255,
        help_text=_("Tags can be used by the project to filter sites"))
    base = models.ForeignKey('self',
        null=True, on_delete=models.CASCADE,
        help_text=_("The site is a derivative of this parent."))

    # Database connection
    db_name = models.SlugField(max_length=255, null=True,
        help_text=_("name of database to connect to for the site"))
    db_host = models.CharField(max_length=255, null=True, blank=True,
        validators=[HOST_VALIDATOR],
        help_text=_("host to connect to the database"))
    db_port = models.IntegerField(null=True, blank=True,
        help_text=_("port to connect to the database host"))
    db_host_user = models.CharField(max_length=128, null=True, blank=True,
        help_text=_("username authorized to connect to the database"))
    db_host_password = models.CharField(_('Password'), max_length=128,
        null=True, blank=True,
        help_text=_("password to authenticate user connecting to the database"))

    # SMTP connection
    email_default_from = models.EmailField(null=True, blank=True)
    email_host = models.CharField(max_length=255, null=True, blank=True,
        validators=[HOST_VALIDATOR],
        help_text=_("host to connect to the SMTP server"))
    email_port = models.IntegerField(null=True, blank=True,
        help_text=_("port to connect to the SMTP server"))
    email_host_user = models.CharField(max_length=128, null=True, blank=True,
        help_text=_("username authorized to send e-mails on the SMTP server"))
    email_host_password = models.CharField(_('Password'), max_length=128,
        null=True, blank=True,
        help_text=_("password to authenticate the user with the SMTP server"))

    class Meta:
        swappable = 'MULTITIER_SITE_MODEL'
        abstract = True

    def __str__(self):
        return str(self.slug)

    def as_base(self):
        """
        Returns either the base site or ``self`` if no base exists.
        """
        if self.base_id:
            return self.base
        return self

    def as_subdomain(self):
        return self.slug if self.slug != settings.DEFAULT_SITE else ""

    @property
    def printable_name(self):
        if self.domain:
            return self.domain
        return self.slug

    def get_templates(self):
        """
        Returns a list of candidate themes.
        """
        return [self.slug]

    def get_template_dirs(self):
        """
        Returns a list of candidate search paths for templates.
        """
        return [safe_join(theme_dir, theme, 'templates')
                for theme_dir in settings.THEMES_DIRS
                    for theme in self.get_templates()]

    def add_tags(self, tags):
        try:
            extra = json.loads(self.extra)
        except ValueError:
            extra = {}
        extra.update({
            'tags': tags + [
                tag for tag in extra.get('tags', []) if tag not in tags]})
        self.extra = json.dumps(extra)

    def remove_tags(self, tags):
        try:
            extra = json.loads(self.extra)
        except ValueError:
            extra = {}
        extra.update({
            'tags': [tag for tag in extra.get('tags', []) if tag not in tags]})
        self.extra = json.dumps(extra)

    @property
    def has_custom_connection(self):
        return (self.email_host_user or self.email_host_password or
            self.email_host or self.email_port)

    def get_email_connection(self):
        kwargs = {}
        if self.email_host:
            kwargs['host'] = self.email_host
        if self.email_port:
            kwargs['port'] = self.email_port
        if self.email_host_user:
            kwargs['username'] = self.email_host_user
        if self.email_host_password:
            kwargs['password'] = self.get_email_host_password()
        return get_connection_base(**kwargs)

    def get_from_email(self):
        if self.email_default_from:
            return self.email_default_from
        if self.email_host_user and '@' in self.email_host_user:
            return self.email_host_user
        return settings.DEFAULT_FROM_EMAIL

    def set_email_host_password(self, raw_password, passphrase=None):
        if not passphrase:
            passphrase = settings.SECRET_KEY
        encrypted = encrypt(raw_password, passphrase=passphrase)
        # b64encode will return `bytes` (Py3) but Django 2.0 is expecting
        # a `str`, otherwise it wraps those `bytes` into a b'***'.
        # Note that Django 1.11 will add those `bytes` "as-is".
        if not isinstance(encrypted, six.string_types):
            self.email_host_password = encrypted.decode('ascii')
        else:
            self.email_host_password = encrypted

    def get_email_host_password(self, passphrase=None):
        if not passphrase:
            passphrase = settings.SECRET_KEY
        return decrypt(self.email_host_password, passphrase=passphrase)


@python_2_unicode_compatible
class Site(BaseSite):

    def __str__(self):
        return str(self.slug)


def get_site_or_none(subdomain):
    """
    Returns a ``Site`` instance based on its subdomain while prefering
    ``Site`` with an explicit domain.
    If no Site could be found, then returns ``None``.
    """
    return get_site_model().objects.filter(
        slug=subdomain).order_by('domain', '-pk').first()



# Copyright (c) 2020, Djaodjin Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import absolute_import

import re

from django.core.exceptions import ImproperlyConfigured
from django.urls import base
from django.utils.datastructures import MultiValueDict
from django.utils.regex_helper import normalize

from .compat import (RegexURLResolver as DjangoRegexURLResolver,
    RegexURLPattern as DjangoRegexURLPattern, lru_cache, six)
from .thread_locals import get_current_site


class SitePrefixPattern(object):

    def __init__(self):
        self.converters = {}

    @staticmethod
    def _get_path_prefix():
        path_prefix = ""
        current_site = get_current_site()
        if current_site and current_site.path_prefix:
            path_prefix = "%s/" % current_site.path_prefix
        return path_prefix

    @property
    def regex(self):
        # This is only used by reverse() and cached in _reverse_dict.
        return re.compile(self._get_path_prefix(), re.UNICODE)

    def match(self, path):
        path_prefix = self._get_path_prefix()
        if path.startswith(path_prefix):
            return path[len(path_prefix):], (), {}
        return None

    @staticmethod
    def check():
        return []

    def describe(self):
        return "'{}'".format(self)

    def __str__(self):
        return self._get_path_prefix()


class BaseRegexURLResolver(DjangoRegexURLResolver):
    """
    A URL resolver that always matches the active organization code
    as URL prefix.
    """
    def __init__(self, regex, urlconf_name, *args, **kwargs):
        super(BaseRegexURLResolver, self).__init__(
            regex, urlconf_name, *args, **kwargs)

    @staticmethod
    def _get_path_prefix():
        path_prefix = "_"
        current_site = get_current_site()
        if current_site and current_site.path_prefix:
            path_prefix = current_site.path_prefix
        return path_prefix

    # Implementation Note:
    # Copy/Pasted `RegexURLResolver._populate` here because that was the only
    # way to override `language_code = get_language()` to use a dynamic path
    # prefix `path_prefix = self._get_path_prefix()`.
    def _populate(self):
        # Short-circuit if called recursively in this thread to prevent
        # infinite recursion. Concurrent threads may call this at the same
        # time and will need to continue, so set 'populating' on a
        # thread-local variable.
        #pylint:disable=protected-access,too-many-locals,too-many-nested-blocks
        if getattr(self._local, 'populating', False):
            return
        try:
            self._local.populating = True
            lookups = MultiValueDict()
            namespaces = {}
            apps = {}
            path_prefix = self._get_path_prefix()
            for url_pattern in reversed(self.url_patterns):
                if isinstance(url_pattern, DjangoRegexURLPattern):
                    self._callback_strs.add(url_pattern.lookup_str)
                # could be RegexURLPattern.regex or RegexURLResolver.regex here.
                p_pattern = url_pattern.regex.pattern
                if p_pattern.startswith('^'):
                    p_pattern = p_pattern[1:]
                if isinstance(url_pattern, DjangoRegexURLResolver):
                    if url_pattern.namespace:
                        namespaces[url_pattern.namespace] = (
                            p_pattern, url_pattern)
                        if url_pattern.app_name:
                            apps.setdefault(
                                url_pattern.app_name, []).append(
                                    url_pattern.namespace)
                    else:
                        parent_pat = url_pattern.regex.pattern
                        for name in url_pattern.reverse_dict:
                            for _, pat, defaults \
                                in url_pattern.reverse_dict.getlist(name):
                                new_matches = normalize(parent_pat + pat)
                                lookups.appendlist(
                                    name,
                                    (
                                        new_matches,
                                        p_pattern + pat,
                                        dict(defaults,
                                             **url_pattern.default_kwargs),
                                    )
                                )
                        for namespace, (prefix, sub_pattern) \
                            in url_pattern.namespace_dict.items():
                            namespaces[namespace] = (
                                p_pattern + prefix, sub_pattern)
                        for app_name, namespace_list in \
                                url_pattern.app_dict.items():
                            apps.setdefault(app_name, []).extend(namespace_list)
                    if not getattr(url_pattern._local, 'populating', False):
                        url_pattern._populate()
                    self._callback_strs.update(url_pattern._callback_strs)
                else:
                    bits = normalize(p_pattern)
                    lookups.appendlist(url_pattern.callback, (
                        bits, p_pattern, url_pattern.default_args))
                    if url_pattern.name is not None:
                        lookups.appendlist(url_pattern.name, (
                            bits, p_pattern, url_pattern.default_args))
            self._reverse_dict[path_prefix] = lookups
            self._namespace_dict[path_prefix] = namespaces
            self._app_dict[path_prefix] = apps
            self._populated = True
        finally:
            self._local.populating = False

    @property
    def reverse_dict(self):
        path_prefix = self._get_path_prefix()
        if path_prefix not in self._reverse_dict:
            self._populate()
        return self._reverse_dict[path_prefix]

    @property
    def namespace_dict(self):
        path_prefix = self._get_path_prefix()
        if path_prefix not in self._namespace_dict:
            self._populate()
        return self._namespace_dict[path_prefix]

    @property
    def app_dict(self):
        path_prefix = self._get_path_prefix()
        if path_prefix not in self._app_dict:
            self._populate()
        return self._app_dict[path_prefix]


try:

    from .urlresolvers_py3_django2 import RegexURLResolver

except (ImportError, SyntaxError): # <= Django2, Python 2

    class RegexURLResolver(BaseRegexURLResolver):
        pass

except ModuleNotFoundError: #pylint:disable=bad-except-order
    # <= Django2, Python 3
    class RegexURLResolver(BaseRegexURLResolver):
        pass


class SiteRegexURLResolver(RegexURLResolver):
    """
    A URL resolver that always matches the active organization code
    as URL prefix.
    """
    def __init__(self, regex, urlconf_name, *args, **kwargs):
        super(SiteRegexURLResolver, self).__init__(
            regex, urlconf_name, *args, **kwargs)

    @property
    def regex(self):
        path_prefix = self._get_path_prefix()
        if path_prefix != '_':
            # site will be None when 'manage.py show_urls' is invoked.
            return re.compile('^%s/' % path_prefix, re.UNICODE)
        return re.compile('^', re.UNICODE)


def site_patterns(*args):
    """
    Adds the live organization prefix to every URL pattern within this
    function. This may only be used in the root URLconf, not in an included
    URLconf.
    """
    pattern_list = args
    return [SiteRegexURLResolver(SitePrefixPattern(), pattern_list)]


try:
    from django.urls.resolvers import RegexPattern

    @lru_cache(maxsize=None)
    def get_resolver(urlconf=None):
        if urlconf is None:
            from django.conf import settings
            urlconf = settings.ROOT_URLCONF
        return RegexURLResolver(RegexPattern(r'^/'), urlconf)

    def url_sites(regex, view, kwargs=None, name=None, prefix=''):
        #pylint:disable=unused-argument
        if not view:
            raise ImproperlyConfigured(
                'Empty URL pattern view name not permitted (for pattern %r)'
                % regex)
        if isinstance(view, (list, tuple)):
            # For include(...) processing.
            pattern = RegexPattern(regex, is_endpoint=False)
            urlconf_module, app_name, namespace = view
            return RegexURLResolver(
                pattern,
                urlconf_module,
                kwargs,
                app_name=app_name,
                namespace=namespace,
            )
        if callable(view):
            pattern = RegexPattern(regex, name=name, is_endpoint=True)
            return DjangoRegexURLPattern(pattern, view, kwargs, name)
        raise TypeError('view must be a callable or a list/tuple'\
            ' in the case of include().')

except ImportError:
    def url_sites(regex, view, kwargs=None, name=None, prefix='',
            pattern=DjangoRegexURLPattern, resolver=RegexURLResolver):
        """
        Modified `django.conf.urls.url` with allows to specify custom
        RegexURLPattern and RegexURLResolver classes.
        """
        #pylint:disable=too-many-arguments
        if isinstance(view, (list, tuple)):
            # For include(...) processing.
            return resolver(regex, view[0], kwargs, *view[1:])

        if isinstance(view, six.string_types):
            if not view:
                raise ImproperlyConfigured('Empty URL pattern view'\
                    ' name not permitted (for pattern %r)' % regex)
            if prefix:
                view = prefix + '.' + view
        return pattern(regex, view, kwargs, name)

    @lru_cache(maxsize=None)
    def get_resolver(urlconf=None):
        if urlconf is None:
            from django.conf import settings
            urlconf = settings.ROOT_URLCONF
        return RegexURLResolver(r'^/', urlconf)


# Severe monkey patching without which calling the top level resolver
# `_reverse_with_prefix` method is not updating caches for *path_prefix*.
base.get_resolver = get_resolver



# Copyright (c) 2017, Djaodjin Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from django.shortcuts import get_object_or_404

from . import settings
from .compat import get_model_class, import_string
from .thread_locals import cache_provider_db, get_current_site
from .utils import get_site_model
from .compat import urlparse


def is_localhost(hostname):
    return bool(
        hostname.startswith('localhost') or hostname.startswith('127.0.0.1'))


def build_absolute_uri(request, location='/', site=None,
                       with_scheme=True, force_subdomain=False):
    """
    Builds an absolute URL for path *location* on *site*.

    It is sometimes useful to force subdomains URL. For example when DNS is not
    yet setup and we still want to be able to access the newly created site.

    If *force_subdomain* is ``True``, the site.domain field is not
    used, regardless of its value (valid or null). By default site.domain
    is used when present.
    """
    parts = urlparse(location)
    if parts.scheme and parts.netloc:
        # If we already have an absolute URI then we have not processing
        # to do and just return it "as is".
        return location

    if site is None:
        site = request.site if hasattr(request, 'site') else get_current_site()
    else:
        site_model = get_site_model()
        if not isinstance(site, site_model):
            try:
                site = site_model.objects.get(slug=site)
            except site_model.DoesNotExist:
                site = None
    actual_domain = settings.DEFAULT_DOMAIN
    if request:
        actual_domain = request.get_host()
    if site:
        if site.domain and not force_subdomain:
            actual_domain = site.domain
        else:
            subdomain = site.as_subdomain()
            is_path_prefix = site.is_path_prefix
            force_path_prefix = False
            base_domain = settings.DEFAULT_DOMAIN
            if request:
                base_domain = request.get_host()
                if is_localhost(base_domain):
                    force_path_prefix = True
            if subdomain:
                if not (force_path_prefix or is_path_prefix):
                    actual_domain = '%s.%s' % (subdomain, base_domain)
                elif not location.startswith('/%s/' % subdomain):
                    # In local development, we force use of path prefixes.
                    # At the same time we don't want to double the path prefix
                    # when it was already added by ``reverse()``.
                    actual_domain = '%s/%s' % (base_domain, subdomain)

    result = "%(domain)s%(path)s" % {'domain': actual_domain, 'path': location}
    if with_scheme:
        scheme = 'http'
        if request and not is_localhost(result):
            scheme = request.scheme
        result = '%s://%s' % (scheme, result)
    return result


class AccountMixin(object):

    account_url_kwarg = settings.ACCOUNT_URL_KWARG

    @property
    def account(self):
        if not hasattr(self, '_account'):
            if settings.ACCOUNT_GET_CURRENT:
                self._account = import_string(settings.ACCOUNT_GET_CURRENT)(
                    self.kwargs.get(self.account_url_kwarg))
            else:
                self._account = get_object_or_404(
                    get_model_class(settings.ACCOUNT_MODEL,
                        "MULTITIER['ACCOUNT_MODEL']"),
                    slug=self.kwargs.get(self.account_url_kwarg, None))
        return self._account


class SiteMixin(AccountMixin):
    """
    Returns a ``Site`` from a URL or the default site if none can be found.
    """
    site_url_kwarg = 'site'

    @property
    def site(self):
        #pylint: disable=access-member-before-definition
        if not hasattr(self, '_site') or self._site is None:
            if self.site_url_kwarg in self.kwargs:
                self._site = get_object_or_404(
                    get_site_model(), slug=self.kwargs.get(self.site_url_kwarg),
                    account=self.account)
            else:
                self._site = get_current_site().db_object
            cache_provider_db(
                self._site.db_name if self._site.db_name else 'default',
                db_host=self._site.db_host, db_port=self._site.db_port)
        return self._site

    def get_actual_domain(self):
        return build_absolute_uri(
            self.request, site=self.site, with_scheme=False)

    def get_absolute_uri(self, location=''):
        return build_absolute_uri(self.request, location, site=self.site)
